/*
 * File:	baikal_misc.c
 * Description:	Miscellaneous initialization routines for Baikal-T1 SoC
 *
 * Copyright 2019-2020, Baikal Electronics JSC
 */

#include <common.h>
#include <dm.h>
#include <clk.h>
#include <fdt_support.h>
#include <asm/baikal_hw.h>
#include <asm/io.h>
#include <asm/pmu.h>

#define pow2(n)			(1ULL<<(n))   /* power of two */
#define to_bytes(x)		(((u64)(x))<<20)  /* convert to bytes */
#define to_megabytes(x)	(((u64)(x))>>20)  /* convert to mega_bytes */

DECLARE_GLOBAL_DATA_PTR;

unsigned long baikal_clk_freq = CONFIG_SYS_CLK_FREQ;

/*
 * Check if XPA feature is enabled. If so, length of memory
 * availiable in Linux can surpass 32-bit length.
 * To check whether XPA feature is enabled read the
 * MIPS_CP0_TLB_PG_GRAIN ($5, 1) register from Coprocessor0.
 */
static u64 baikal_address_space (void)
{
	u32 page_grain_register;
	asm volatile("mfc0	%0, $5, 1" : "=r" (page_grain_register));
	if (page_grain_register)
		return pow2(40);   /* 40-bit length, 0x100_0000_0000, 1TB */
	else
		return pow2(32);   /* 32-bit length, 0x001_0000_0000, 4GB */
}

static int baikal_set_freq(unsigned long freq)
{
	struct clk clk;
	struct udevice *dev;
	int rc;

	rc = uclass_get_device_by_name(UCLASS_CLK, BAIKAL_CPU_CLK_NAME, &dev);
	if (rc) {
		printf("Can't get clk %s (%d)\n", BAIKAL_CPU_CLK_NAME, rc);
		return rc;
	}
	rc = clk_get_by_index(dev, 0, &clk);
	if (rc) {
		printf("Can't get clk (%d)\n", rc);
		return rc;
	}
	rc = clk_set_rate(&clk, freq);
	if (!rc) {
		baikal_clk_freq = freq;
		gd->cpu_clk = freq;
		printf("CPU frequency set to %luMHz\n", freq / 1000000);
	} else {
		printf("Can't set freq to %lu (%d)\n", freq, rc);
	}

	return rc;
}

static int on_cpufreq(const char *name, const char *value, enum env_op op, int flags)
{
	unsigned long freq;

	freq = strtoul(value, NULL, 10);
	if (freq < 100 || freq > 1500) {
		printf("Frequency %lu out of range\n", freq);
		return -EINVAL;
	}
	freq *= 1000000;

	return baikal_set_freq(freq);
}
U_BOOT_ENV_CALLBACK(cpufreq, on_cpufreq);

#ifdef CONFIG_BOARD_EARLY_INIT_R
int board_early_init_r(void)
{
	ulong val;

	/* switch boot controller into SPI mode */
	val = readl((void __iomem *)BAIKAL_BOOT_BASE);
	val |= 0x100;
	writel(val, (void __iomem *)BAIKAL_BOOT_BASE);

	/* Enable whatever clocks not enabled by reset */
#ifndef CONFIG_CLK
	/* SATA: */
	val = READ_PMU_REG(BK_PMU_AXI_SATA_CTL);
	val |= 1;
	WRITE_PMU_REG(BK_PMU_AXI_SATA_CTL, val);

	val = READ_PMU_REG(BK_PMU_SATA_REF_CTL);
	val |= 1;
	WRITE_PMU_REG(BK_PMU_SATA_REF_CTL, val);

	/* USB: */
	val = READ_PMU_REG(BK_PMU_AXI_USB_CTL);
	val |= 1;
	WRITE_PMU_REG(BK_PMU_AXI_USB_CTL, val);

	val = READ_PMU_REG(BK_PMU_USB_CTL);
	val |= 1;
	WRITE_PMU_REG(BK_PMU_USB_CTL, val);

	val = READ_PMU_REG(BK_PMU_AXI_GMAC1_CTL);
	val |= 1; /* enable */
	WRITE_PMU_REG(BK_PMU_AXI_GMAC1_CTL, val);

	val = READ_PMU_REG(BK_PMU_AXI_GMAC2_CTL);
	val |= 1; /* enable */
	WRITE_PMU_REG(BK_PMU_AXI_GMAC2_CTL, val);
#endif

	return 0;
}
#endif /* CONFIG_BOARD_EARLY_INIT_R */

void board_preboot_os(void)
{
	u32 val;

	/* restore boot controller to "ROM" state */
	val = readl((void __iomem *)BAIKAL_BOOT_BASE);
	val &= ~0x100;
	writel(val, (void __iomem *)BAIKAL_BOOT_BASE);
}

/*
* memory map: [[bank0],[bank1]]
*/
int ft_board_setup(void *blob, bd_t *bd)
{
	u64 start[2];
	u64 size[2];
	u64 total = baikal_address_space();  /* 4GB,1TB */
	u64 high = to_bytes(gd->arch.highmem_size);

	/* bank start */
	start[0] = DDR_BANK0_START;
	start[1] = DDR_BANK1_START;

	/* bank size */
	size[0] = DDR_BANK0_SIZE;
	if (size[0] > start[1] - start[0])
		size[0] = start[1] - start[0];

	size[1] = high;
	if (size[1] > total - start[1])
		size[1] = total - start[1];

	/* fixup */
	return fdt_fixup_memory_banks(blob, start, size, 2);
}
