// SPDX-License-Identifier: GPL-2.0
/*
 * File:	board.c
 * Description:	BFK3 Baikal-T1 based board initialization
 * Authors:
 *
 * Copyright 2019-2020, Baikal Electronics JSC
 */

#include <common.h>
#include <scsi.h>
#include <usb.h>
#include <stdio_dev.h>
#include <asm/gpio.h>
#include <spi.h>
#include <spi_flash.h>
#include <env.h>

DECLARE_GLOBAL_DATA_PTR;

#define BOOT_FLASH_ID_LEN	20
#define CONFIG_BAIKAL_MAC_BASE	0x4CA515000000ULL

#define ETH_ID_1G_1	1
#define ETH_ID_1G_2	2
#define ETH_ID_10G	0

void calc_and_set_macaddr(int serial)
{
	u64 macaddr;
	u8* b = (u8*)&macaddr;
	uint8_t enetaddr[6];

	macaddr = CONFIG_BAIKAL_MAC_BASE | ((serial << 2) & 0xFFFFFC);
	for (int i = 0; i < 6; i++) enetaddr[i] = b[5 - i];

	/* Ethernet 10G (eth0) */
	enetaddr[5] = b[0] | ETH_ID_10G;
	eth_env_set_enetaddr("ethaddr", enetaddr);

	/* Ethernet 1G (eth1) */
	enetaddr[5] = b[0] | ETH_ID_1G_1;
	eth_env_set_enetaddr("eth1addr", enetaddr);

	/* Ethernet 1G (eth2) */
	enetaddr[5] = b[0] | ETH_ID_1G_2;
	eth_env_set_enetaddr("eth2addr", enetaddr);
}

unsigned long get_spi_chip_id(void) {
	int spi_get_bus_error, spi_claim_error, spi_read_error;

	struct spi_slave *slave;
	struct udevice *bus;

	uint8_t data[BOOT_FLASH_ID_LEN + 1];
	const char cmd[BOOT_FLASH_ID_LEN + 1] = {0x9F};

	char flash_chip_id[BOOT_FLASH_ID_LEN * 2 + 1];

	flash_chip_id[BOOT_FLASH_ID_LEN * 2] = '\0';

	spi_get_bus_error = spi_get_bus_and_cs(CONFIG_SF_DEFAULT_BUS, CONFIG_SF_DEFAULT_CS, CONFIG_SF_DEFAULT_SPEED, CONFIG_SF_DEFAULT_MODE, NULL, 0, &bus, &slave);

	if (spi_get_bus_error) {
		printf("Error: Failed to get spi bus (%d)\n", spi_get_bus_error);
		return env_get_ulong("cpu_serial", 10, 0);
	}
	else {
		spi_claim_error = spi_claim_bus(slave);
		if (spi_claim_error) {
				printf("Error: Failed to claim spi bus (%d)\n", spi_claim_error);
				return env_get_ulong("cpu_serial", 10, 0);
		}
		else {
			spi_read_error = spi_xfer(slave, 168, cmd, data, SPI_XFER_BEGIN | SPI_XFER_END);
			spi_release_bus(slave);

			if (spi_read_error) {
				printf("Error: Failed to read boot flash ID (%d)\n", spi_read_error);
				return env_get_ulong("cpu_serial", 10, 0);
			}
			else {
				for (int i = 1; i < BOOT_FLASH_ID_LEN + 1; i++)
					sprintf(flash_chip_id + (i - 1) * 2, "%02x", data[i]);
				printf("RomID: %s\n", flash_chip_id);

				return simple_strtol(flash_chip_id + (BOOT_FLASH_ID_LEN - 6), NULL, 16);
			}
		}
	}
}

#ifdef CONFIG_BOARD_LATE_INIT
int board_late_init(void)
{
	char *stdio_name;
	unsigned long brd_serial_id;

	debug("bfk3: board_late_init\n");

	brd_serial_id = get_spi_chip_id();
	calc_and_set_macaddr(brd_serial_id);

#ifdef CONFIG_DM_PCI
	pci_init();
#ifdef CONFIG_DM_VIDEO
	stdio_name = env_get("stdout");
	if (stdio_name && strstr(stdio_name, "vidconsole"))
		env_set("stdout", stdio_name);
#endif
#endif
#ifdef CONFIG_DM_USB
	usb_init();
#ifdef CONFIG_USB_KEYBOARD
	stdio_name = env_get("stdin");
	if (stdio_name && strstr(stdio_name, "usbkbd"))
		env_set("stdin", stdio_name);
#endif
#endif
#ifdef CONFIG_DM_SCSI
	scsi_scan(false);
#endif

	return 0;
}
#endif
