// SPDX-License-Identifier: GPL-2.0+
/*
 * Copyright (C) 2013 Altera Corporation <www.altera.com>
 */

#include <common.h>
#ifdef CONFIG_WDT
#include <dm.h>
#include <wdt.h>
#include <clk.h>
#else
#include <watchdog.h>
#endif
#include <asm/io.h>
#include <asm/utils.h>

#define DW_WDT_CR	0x00
#define DW_WDT_TORR	0x04
#define DW_WDT_CRR	0x0C

#define DW_WDT_CR_EN_OFFSET	0x00
#define DW_WDT_CR_RMOD_OFFSET	0x01
#define DW_WDT_CR_RMOD_VAL	0x00
#define DW_WDT_CRR_RESTART_VAL	0x76

struct dw_wdt_priv {
	void __iomem *regs;
	unsigned long clk_rate;
};

#ifndef CONFIG_WDT
static struct dw_wdt_priv wdt_priv;
#endif

/*
 * Set the watchdog time interval.
 * Counter is 32 bit.
 */
static int designware_wdt_settimeout(struct dw_wdt_priv *priv, unsigned int timeout)
{
	signed int i;

	/* calculate the timeout range value */
	i = (log_2_n_round_up(timeout * (priv->clk_rate / 1000))) - 16;
#ifdef CONFIG_TARGET_BAIKAL_MIPS
	i += 4;
#endif
	if (i > 15)
		i = 15;
	if (i < 0)
		i = 0;

	writel((i | (i << 4)), (priv->regs + DW_WDT_TORR));
	return 0;
}

static void designware_wdt_enable(struct dw_wdt_priv *priv)
{
	writel(((DW_WDT_CR_RMOD_VAL << DW_WDT_CR_RMOD_OFFSET) |
	      (0x1 << DW_WDT_CR_EN_OFFSET)),
	      (priv->regs + DW_WDT_CR));
}

static unsigned int designware_wdt_is_enabled(struct dw_wdt_priv *priv)
{
	unsigned long val;
	val = readl((priv->regs + DW_WDT_CR));
	return val & 0x1;
}

#if defined(CONFIG_HW_WATCHDOG) && !defined(CONFIG_WDT)
void hw_watchdog_reset(void)
{
	struct dw_wdt_priv *priv = &wdt_priv;

	if (designware_wdt_is_enabled(priv))
		/* restart the watchdog counter */
		writel(DW_WDT_CRR_RESTART_VAL,
		       (priv->regs + DW_WDT_CRR));
}

void hw_watchdog_init(void)
{
	struct dw_wdt_priv *priv;

	wdt_priv.regs = (void *)CONFIG_DW_WDT_BASE,
	wdt_priv.clk_rate = CONFIG_DW_WDT_CLOCK_KHZ * 1000,

	priv = &wdt_priv;
	/* reset to disable the watchdog */
	hw_watchdog_reset();
	/* set timer in miliseconds */
	designware_wdt_settimeout(priv, CONFIG_WATCHDOG_TIMEOUT_MSECS);
	/* enable the watchdog */
	designware_wdt_enable(priv);
	/* reset the watchdog */
	hw_watchdog_reset();
}
#endif

#ifdef CONFIG_WDT
static int dw_wdt_start(struct udevice *dev, u64 timeout, ulong flags)
{
	struct dw_wdt_priv *priv = dev_get_priv(dev);

	designware_wdt_settimeout(priv, timeout);
	designware_wdt_enable(priv);

	return 0;
}

static int dw_wdt_reset(struct udevice *dev)
{
	struct dw_wdt_priv *priv = dev_get_priv(dev);

	if (designware_wdt_is_enabled(priv))
		/* restart the watchdog counter */
		writel(DW_WDT_CRR_RESTART_VAL,
		       (priv->regs + DW_WDT_CRR));

	return 0;
}

static int dw_wdt_expire_now(struct udevice *dev, ulong flags)
{
	struct dw_wdt_priv *priv = dev_get_priv(dev);

	designware_wdt_settimeout(priv, 1);
	designware_wdt_enable(priv);
	/* restart the watchdog counter */
	writel(DW_WDT_CRR_RESTART_VAL, (priv->regs + DW_WDT_CRR));

	return 0;
}

static struct wdt_ops dw_wdt_ops = {
	.reset = dw_wdt_reset,
	.start = dw_wdt_start,
	.expire_now = dw_wdt_expire_now,
};

static const struct udevice_id dw_wdt_ids[] = {
	{ .compatible = "snps,dw-wdt", },
	{ /* sentinel */ }
};

static int dw_wdt_probe(struct udevice *dev)
{
	struct dw_wdt_priv *priv = dev_get_priv(dev);
	struct clk clk;
	int ret;

	priv->regs = devfdt_map_physmem(dev, 0x10);
	if (!priv->regs)
		return -EINVAL;

#ifdef CONFIG_CLK
	ret = clk_get_by_index(dev, 0, &clk);
	if (!ret)
		priv->clk_rate = clk_get_rate(&clk);
	else
		return -EINVAL;
#else
	priv->clk_rate = CONFIG_DW_WDT_CLOCK_KHZ * 1000;
#endif

	return 0;
}

U_BOOT_DRIVER(wdt_dw) = {
	.name = "wdt_dw",
	.id = UCLASS_WDT,
	.of_match = dw_wdt_ids,
	.ops = &dw_wdt_ops,
	.priv_auto_alloc_size = sizeof(struct dw_wdt_priv),
	.probe = dw_wdt_probe,
};
#endif

